''' DEPRECATED element definitions, based on dictionaries instead
    of classes, translated into their newer class counterparts.
'''

from functools import partial

from .. import elements as elm

RES = elm.Resistor
RES_VAR = elm.ResistorVar
RBOX = elm.ResistorIEC
POT = elm.Potentiometer
CAP = elm.Capacitor
CAP_P = partial(elm.Capacitor, polar=True)
CAP2 = elm.Capacitor2
CAP2_P = partial(elm.Capacitor2, polar=True)
CAP_VAR = elm.CapacitorVar
INDUCTOR = elm.Inductor
INDUCTOR2 = elm.Inductor2
SOURCE = elm.Source
SOURCE_V = elm.SourceV
SOURCE_I = elm.SourceI
SOURCE_SIN = elm.SourceSin
SOURCE_CONT = elm.SourceControlled
SOURCE_CONT_I = elm.SourceControlledI
SOURCE_CONT_V = elm.SourceControlledV
BATT_CELL = elm.BatteryCell
BATTERY = elm.Battery
METER_V = elm.MeterV
METER_I = elm.MeterI
METER_OHM = elm.MeterOhm
XTAL = elm.Crystal
DIODE = elm.Diode
DIODE_F = partial(elm.Diode, fill=True)
SCHOTTKY = elm.Schottky
SCHOTTKY_F = partial(elm.Schottky, fill=True)
DIODE_TUNNEL = elm.DiodeTunnel
DIODE_TUNNEL_F = partial(elm.DiodeTunnel, fill=True)
ZENER = elm.Zener
ZENER_F = partial(elm.Zener, fill=True)
LED = elm.LED  # Same name!
LED2 = elm.LED2
PHOTODIODE = elm.Photodiode
DIAC = elm.Diac
DIAC_F = partial(elm.Diac, fill=True)
TRIAC = elm.Triac
TRIAC_F = partial(elm.Triac, fill=True)
SCR = elm.SCR
SCR_F = partial(elm.SCR, fill=True)
MEMRISTOR = elm.Memristor
MEMRISTOR2 = elm.Memristor2
JJ = elm.Josephson
FUSE = elm.Fuse
DOT_OPEN = partial(elm.Dot, fill='white')
DOT = elm.Dot
ARROWHEAD = elm.Arrowhead
LINE = elm.Line
ARROW = elm.Arrow
ARROW_DOUBLE = partial(elm.Arrow, double=True)
ARROWLINE = elm.CurrentLabel
ARROW_I = elm.CurrentLabel
LINEDOT = elm.LineDot
LINE_DOT = elm.LineDot
LINE_DOT_DOUBLE = partial(elm.LineDot, double=True)
LINE_DOT_OPEN = partial(elm.LineDot, fill='white')
LINE_DOT_OPEN_DOUBLE = partial(elm.LineDot, fill='white', double=True)
ELLIPSIS = elm.DotDotDot
GAP_LABEL = elm.Gap
GAP = elm.Gap
LABEL = elm.Label
GND = elm.Ground
GND_SIG = elm.GroundSignal
GND_CHASSIS = elm.GroundChassis
ANT = elm.Antenna
VSS = elm.Vss
VDD = elm.Vdd
coax = COAX = elm.Coax
triax = TRIAX = elm.Triax
OPAMP_NOSIGN = partial(elm.Opamp, sign=False)
OPAMP = elm.Opamp
NFET = elm.NFet
NFET4 = partial(elm.NFet, bulk=True)
PFET = elm.PFet
PFET4 = partial(elm.PFet, bulk=True)
JFET_N = elm.JFetN
JFET_N_C = partial(elm.JFetN, circle=True)
JFET_P = elm.JFetP
JFET_P_C = partial(elm.JFetP, circle=True)
BJT = elm.Bjt
BJT_NPN = elm.BjtNpn
BJT_NPN_C = partial(elm.BjtNpn, circle=True)
BJT_PNP = elm.BjtPnp
BJT_PNP_C = partial(elm.BjtPnp, circle=True)
BJT_PNP_2C = elm.BjtPnp2c
SWITCH_SPST = elm.Switch
SWITCH_SPST_OPEN = partial(elm.Switch, action='open')
SWITCH_SPST_CLOSE = partial(elm.Switch, action='close')
SWITCH_SPDT = elm.SwitchSpdt
SWITCH_SPDT_OPEN = partial(elm.SwitchSpdt, action='open')
SWITCH_SPDT_CLOSE = partial(elm.SwitchSpdt, action='close')
SWITCH_SPDT2 = elm.SwitchSpdt
SWITCH_SPDT2_OPEN = partial(elm.SwitchSpdt2, action='open')
SWITCH_SPDT2_CLOSE = partial(elm.SwitchSpdt2, action='close')
BUTTON = elm.Button
BUTTON_NC = partial(elm.Button, nc=True)
SPEAKER = elm.Speaker
MIC = elm.Mic
LAMP = elm.Lamp
MOTOR = elm.Motor
transformer = elm.Transformer


def ic(*pins, **kwargs):
    ''' Define an integrated circuit element

        Parameters
        ----------
        *pins: dict
            A dictionary defining each input/output pin.
            Each dictionary may contain the following
            optional keys:

            name: string
                Signal name, labeled inside the IC box.
                If name is '>', a proper clock input triangle
                will be drawn instead of a text label.
            pin: string
                Pin name, labeled outside the IC box
            side: string ['left', 'right', 'top', 'bottom']
                Which side the pin belongs on
            pos: float
                Absolute position as fraction from 0-1 along the
                side. If not provided, pins are evenly spaced along
                the side.
            slot: string
                Position designation for the pin in "X/Y" format
                where X is the pin number and Y the total number
                of pins along the side. Use when missing pins
                are desired with even spacing.
            invert: bool
                Draw an invert bubble outside the pin
            invertradius: float
                Radius of invert bubble
            color: string
                Matplotlib color for label
            rotation: float
                Rotation angle for label (degrees)
            anchorname: string
                Name of anchor at end of pin lead. By default pins
                will have anchors of both the `name` parameter
                and `inXY` where X the side designation
                ['L', 'R', 'T', 'B'] and Y the pin number along
                that side.

        Keyword Arguments
        -----------------
        size: (w, h) tuple
            Size of the IC. If not provided, size is
            automatically determined based on number of
            pins and the pinspacing parameter.
        pinspacing: float
            Smallest distance between pins [1.25]
        edgepadH: float
            Additional distance from edge to first pin on
            vertical sides [.25]
        edgepadW: float
            Additional distance from edge to first pin on
            horizontal sides [.25]
        lblofst: float
            Default offset for (internal) labels [.15]
        plblofst: float
            Default offset for external pin labels [.1]
        leadlen: float
            Length of leads extending from box [.5]
        lblsize: int
            Font size for (internal) labels [14]
        plblsize: int
            Font size for external pin labels [11]
        slant: float
            Degrees to slant top and bottom sides,
            (e.g. for multiplexers) [0]
    '''
    pins = [elm.IcPin(**p) for p in pins]
    return elm.Ic(pins=pins, **kwargs)


def multiplexer(*pins, **kwargs):
    ''' Draw a multiplexer or demultiplexer.

        Parameters
        ----------
        *pins: dict
            List of pin definitions. See IC method.

        Keyword Arguments
        -----------------
        demux: bool
            Draw demultiplexer (opposite slope)
        slant: float
            Angle to slant top/bottom edges
        **kwargs:
            See IC method.
    '''
    slant = kwargs.pop('slant', 25)
    demux = kwargs.pop('demux', False)
    slant = -slant if not demux else slant
    return ic(*pins, slant=slant, **kwargs)



print("import: 'schemdraw'")
import schemdraw

print("import: 'schemdraw.elements'")
import schemdraw.elements

print("import: 'schemdraw.logic'")
import schemdraw.logic

print("import: 'schemdraw.dsp'")
import schemdraw.dsp

print("import: 'schemdraw.flow'")
import schemdraw.flow

print("import: 'schemdraw.backends'")
import schemdraw.backends

print("import: 'schemdraw.parsing'")
import schemdraw.parsing




''' Convert strings into SVG <text> elements, including some basic
    Latex-like math markup and formatting.

    The math only covers the most common cases in encountered
    electrical schematics:
        - Superscripts
        - Subscripts
        - Overline
        - Greek symbols

    This module also approximates the size of text to get bounding boxes
    and rotation centers.
'''

from __future__ import annotations

from xml.etree import ElementTree as ET
import string
import re

from ..util import Point, rotate
from ..types import Halign, Valign, RotationMode


# Conversion from Latex codes to unicode characters.
# Note: keys must include escaping \ for use in regex.
textable = {
    r'\\Gamma': 'Γ',
    r'\\Delta': 'Δ',
    r'\\Theta': 'Θ',
    r'\\Lambda': 'Λ',
    r'\\Xi': 'Ξ',
    r'\\Pi': 'Π',
    r'\\Sigma': 'Σ',
    r'\\Upsilon': 'Υ',
    r'\\Phi': 'Φ',
    r'\\Psi': 'Ψ',
    r'\\Omega': 'Ω',
    r'\\alpha': 'α',
    r'\\beta': 'β',
    r'\\gamma': 'γ',
    r'\\delta': 'δ',
    r'\\epsilon': 'ϵ',
    r'\\varepsilon': 'ε',
    r'\\zeta': 'ζ',
    r'\\eta': 'η',
    r'\\theta': 'θ',
    r'\\vartheta': 'ϑ',
    r'\\iota': 'ι',
    r'\\kappa': 'κ',
    r'\\lambda': 'λ',
    r'\\mu': 'μ', # \u03BC - "Greek small letter mu" (not \uB5 micro sign)
    r'\\nu': 'ν',
    r'\\xi': 'ξ',
    r'\\omicron': 'ο',
    r'\\pi': 'π',
    r'\\varpi': 'ϖ',
    r'\\rho': 'ρ',
    r'\\sigma': 'σ',
    r'\\tau': 'τ',
    r'\\upsilon': 'υ',
    r'\\phi': 'ϕ',
    r'\\varphi': 'φ',
    r'\\chi': 'χ',
    r'\\psi': 'ψ',
    r'\\omega': 'ω',
    r'\\pm': '±',
    r'\\mp': '∓',
    r'\\times': '×',
    r'\\div': '÷',
    r'\\neq': '≠',
    r'\\leq': '≤',
    r'\\geq': '≥',
    r'\\ll': '≪',
    r'\\gg': '≫',
    r'\\subset': '⊂',
    r'\\subseteq': '⊆',
    r'\\sum': '∑',
    r'\\prod': '∏',
    r'\\coprod': '∐',
    r'\\int': '∫',
    r'\\iint': '∬',
    r'\\iiint': '∭',
    r'\\oint': '∮',
    r'\\infty': '∞',
    r'\\nabla': '∇',
    r'\\Re': 'ℜ',
    r'\\Im': 'ℑ',
    r'\\angle': '∠',
    r'\\measuredangle': '∡',
    r'\\mho': '℧',
    r'\\partial': '∂',
    r'\\mathring{A}': 'Å',
    r'\\hbar': 'ℏ',
    r'\^{\\circ}': '°',
    r'\^\{0\}': '⁰',
    r'\^\{1\}': '¹',
    r'\^\{2\}': '²',
    r'\^\{3\}': '³',
    r'\^\{4\}': '⁴',
    r'\^\{5\}': '⁵',
    r'\^\{6\}': '⁶',
    r'\^\{7\}': '⁷',
    r'\^\{8\}': '⁸',
    r'\^\{9\}': '⁹',
    r'_\{0\}': '₀',
    r'_\{1\}': '₁',
    r'_\{2\}': '₂',
    r'_\{3\}': '₃',
    r'_\{4\}': '₄',
    r'_\{5\}': '₅',
    r'_\{6\}': '₆',
    r'_\{7\}': '₇',
    r'_\{8\}': '₈',
    r'_\{9\}': '₉',
    r'\^0': '⁰',
    r'\^1': '¹',
    r'\^2': '²',
    r'\^3': '³',
    r'\^4': '⁴',
    r'\^5': '⁵',
    r'\^6': '⁶',
    r'\^7': '⁷',
    r'\^8': '⁸',
    r'\^9': '⁹',
    r'_0': '₀',
    r'_1': '₁',
    r'_2': '₂',
    r'_3': '₃',
    r'_4': '₄',
    r'_5': '₅',
    r'_6': '₆',
    r'_7': '₇',
    r'_8': '₈',
    r'_9': '₉'}

# Numbers have unique unicode sub/superscripts.
# Some letters do, but unfortunately not all
numsubscripts = {
    '0': '₀',
    '1': '₁',
    '2': '₂',
    '3': '₃',
    '4': '₄',
    '5': '₅',
    '6': '₆',
    '7': '₇',
    '8': '₈',
    '9': '₉'}
numsuperscripts = {
    '0': '⁰',
    '1': '¹',
    '2': '²',
    '3': '³',
    '4': '⁴',
    '5': '⁵',
    '6': '⁶',
    '7': '⁷',
    '8': '⁸',
    '9': '⁹'}


def replacelatex(text: str) -> str:
    ''' Replace latex math codes with unicode equivalents '''
    for k, v in textable.items():
        text = re.sub(k, v, text)
    return text


def mathtextsvg(text: str) -> ET.Element:
    ''' Convert any math string delimited by $..$ into a <tspan> element
        that can be used in <text>

        Args:
            text: The text to convert
    '''
    text = text.replace('>', '%gt;').replace('<', '$lt;')
    tokens = re.split(r'(\$.*?\$)', text)
    svgtext = ''
    for t in tokens:
        if not (t.startswith('$') and t.endswith('$')):
            svgtext += t
            continue

        t = t[1:-1]
        t = replacelatex(t)

        # Find superscripts within {}
        sups = re.split(r'(\^\{.*?\})', t)
        for sup in sups:
            if sup.startswith('^'):
                chrs = sup[2:-1]
                if chrs.isnumeric():
                    chrs = ''.join([numsuperscripts[c] for c in chrs])
                    t = t.replace(sup, chrs)
                else:
                    t = t.replace(sup, f'<tspan baseline-shift="super" font-size="smaller">{chrs}</tspan>')
        # Find superscripts single char
        sups = re.split(r'(\^.)', t)
        for sup in sups:
            if sup.startswith('^'):
                t = t.replace(sup, f'<tspan baseline-shift="super" font-size="smaller">{sup[1]}</tspan>')

        # Find subscripts within {}
        sups = re.split(r'(\_\{.*?\})', t)
        for sup in sups:
            if sup.startswith('_'):
                chrs = sup[2:-1]
                if chrs.isnumeric():
                    chrs = ''.join([numsubscripts[c] for c in chrs])
                    t = t.replace(sup, chrs)
                else:
                    t = t.replace(sup, f'<tspan baseline-shift="sub" font-size="smaller">{chrs}</tspan>')
        # Find subscripts single char
        sups = re.split(r'(\_.)', t)
        for sup in sups:
            if sup.startswith('_'):
                t = t.replace(sup, fr'<tspan baseline-shift="sub" font-size="smaller">{sup[1]}</tspan>')

        # \sqrt
        sups = re.split(r'(\\sqrt{.*})', t)
        for sup in sups:
            if sup.startswith(r'\sqrt{'):
                t = t.replace(sup, fr'√\overline{{{sup[6:-1]}}}')

        # \overline
        sups = re.split(r'(\\overline{.*})', t)
        for sup in sups:
            if sup.startswith(r'\overline{'):
                t = t.replace(sup, f'<tspan text-decoration="overline">{sup[10:-1]}</tspan>')
        svgtext += t
        
    svgtext = '<tspan>' + svgtext + '</tspan>'
    return ET.fromstring(svgtext)


def string_width(st: str, fontsize: float=12, font: str='Arial') -> float:
    ''' Estimate string width based on individual characters

        Args:
            st: string to estimate
            fontsize: font size
            font: font family

        Returns:
            Estimated width of string
    '''
    # adapted from https://stackoverflow.com/a/16008023/13826284
    # The only alternative is to draw the string to an actual canvas

    size = 0  # in milinches
    if 'times' in font.lower() or ('serif' in font.lower() and 'sans' not in font.lower()):
        # Estimates based on Times Roman
        for s in st:
            if s in 'lij:.,;t': size += 47
            elif s in '|': size += 37
            elif s in '![]fI/\\': size += 55
            elif s in '`-(){}r': size += 60
            elif s in 'sJ°': size += 68
            elif s in '"zcae?1': size += 74
            elif s in '*^kvxyμbdhnopqug#$_α' + string.digits: size += 85
            elif s in '#$+<>=~FSP': size += 95
            elif s in 'ELZT': size += 105
            elif s in 'BRC': size += 112
            elif s in 'DAwHUKVXYNQGO': size += 122
            elif s in '&mΩ': size += 130
            elif s in '%': size += 140
            elif s in 'MW@∠': size += 155
            else: size += 60

    else:  # Arial, or other sans fonts
        for s in st:
            if s in 'lij|\' ': size += 37
            elif s in '![]fI.,:;/\\t': size += 50
            elif s in '`-(){}r"': size += 60
            elif s in '*^zcsJkvxyμ°': size += 85
            elif s in 'aebdhnopqug#$L+<>=?_~FZTα' + string.digits: size += 95
            elif s in 'BSPEAKVXY&UwNRCHD': size += 112
            elif s in 'QGOMm%@Ω': size += 140
            elif s in 'W∠': size += 155
            else: size += 75
    return size * 72 / 1000.0 * (fontsize/12)  # to points


def text_approx_size(text: str, font: str='Arial', size: float=16) -> tuple[float, float, float]:
    ''' Get approximate width, height and line spacing of multiline text

        Does not account for descent or ascent of text (for example,
        a "g" going below the baseline), or for superscripts/subscripts.

        Args:
            text: The text to approximate
            font: Font family, either Arial or Times
            size: Font point size

        Returns:
            w: Width of text box, in points
            h: Height of text box, in points
            dy: Spacing between lines of text, in points
    '''
    w = 0.
    h = 0.
    lines = text.splitlines()
    dy = size if len(lines) > 1 else size * 0.8
    for line in lines:
        math = ' '.join(mathtextsvg(line).itertext())  # itertext strips all the tags
        w = max(w, string_width(math, fontsize=size, font=font))
        h += dy
    return w, h, dy


def text_tosvg(text: str, x: float, y: float, font: str='Arial', size: float=16, color: str='black',
               halign: Halign='center', valign: Valign='center',
               rotation: float=0, rotation_mode: RotationMode='anchor',
               testmode: bool=False) -> ET.Element:
    ''' Convert text to svg <text> tag.

        Args:
            text: The text to convert
            x: Anchor x position, points
            y: Anchor y position, points
            font: Font family
            size: Font point size
            color: Font color
            halign: Horizontal alignment (left, center, right)
            valign: Vertical alignment (top, center, bottom)
            rotation: Rotation angle, degrees
            rotation_mode: Rotation mode (anchor, default), see
            Matplotlib https://matplotlib.org/3.1.1/gallery/text_labels_and_annotations/demo_text_rotation_mode.html
            testmode: For testing, draw rectable around text bounding box
                and dot at text anchor

        Returns:
            Text formatted in svg <text> element string.
    '''
    w, h, dy = text_approx_size(text, font=font, size=size)

    textelm = ET.Element('text')
    
    for line in text.splitlines():
        tspan = mathtextsvg(line)
        tspan.set('x', str(x))
        tspan.set('dy', str(dy))
        textelm.append(tspan)

    boxx = x
    if halign == 'center':
        boxx -= w/2
    elif halign == 'right':
        boxx -= w

    ytext = y
    if valign == 'center':
        ytext += h/2
    elif valign == 'top':
        ytext += h
    anchor = {'center': 'middle', 'left': 'start', 'right': 'end'}.get(halign, 'start')

    org = Point((x, y))
    p1 = Point((boxx, ytext)).rotate(-rotation, org)
    p2 = Point((boxx, ytext-h)).rotate(-rotation, org)
    p3 = Point((boxx+w, ytext-h)).rotate(-rotation, org)
    p4 = Point((boxx+w, ytext)).rotate(-rotation, org)

    # Bounding box, unshifted for "default" rotation mode
    pxmin = min(p1[0], p2[0], p3[0], p4[0])
    pxmax = max(p1[0], p2[0], p3[0], p4[0])
    pymin = min(p1[1], p2[1], p3[1], p4[1])
    pymax = max(p1[1], p2[1], p3[1], p4[1])

    dx = dy = 0
    if rotation != 0 and rotation_mode == 'default':
        if halign == 'left':
            dx = x-pxmin
        elif halign == 'center':
            dx = x-(pxmax+pxmin)/2
        else:  # right
            dx = x-pxmax
        if valign == 'top':
            dy = y-pymin
        elif valign == 'center':
            dy = y-(pymax+pymin)/2
        else:  # bottom
            dy = y-pymax
        xform = f'translate({dx} {dy}) rotate({-rotation} {x} {y})'
    else:
        xform = f'rotate({-rotation} {x} {y})'

    textelm.set('x', str(x))
    textelm.set('y', str(ytext-h))
    textelm.set('fill', color)
    textelm.set('font-size', str(size))
    textelm.set('font-family', font)
    textelm.set('text-anchor', anchor)
    textelm.set('transform', xform)

    if testmode:
        pxmin += dx   # Final bounding box
        pxmax += dx
        pymin += dy
        pymax += dy

        topelm = ET.Element('g')  # Put everything in a group
        topelm.append(textelm)
        box = ET.SubElement(topelm, 'rect')
        box.set('x', str(boxx))
        box.set('y', str(ytext-h))
        box.set('width', str(w))
        box.set('height', str(h))
        box.set('style', 'stroke:blue;stroke-width:1;fill:none')
        box.set('transform', xform)
        circ = ET.SubElement(topelm, 'circle')
        circ.set('cx', str(x))
        circ.set('cy', str(y))
        circ.set('r', '1.5')
        circ.set('stroke', 'red')
        circ.set('fill', 'red')
        rect = ET.SubElement(topelm, 'rect')
        rect.set('x', str(pxmin))
        rect.set('y', str(pymin))
        rect.set('width', str(pxmax-pxmin))
        rect.set('height', str(pymax-pymin))
        rect.set('style', 'stroke:red;stroke-width:1;fill:none')
        textelm = topelm

    return textelm



''' Switches and buttons '''

from typing import Optional, Sequence
import math

from .elements import Element, Element2Term, gap
from ..segments import Segment, SegmentCircle, SegmentArc, SegmentPoly
from ..types import Point, ActionType
from ..util import linspace

sw_dot_r = .12


class Switch(Element2Term):
    ''' Toggle Switch

        Args:
            action: action arrow ('open' or 'close')
    '''
    def __init__(self, *d, action: ActionType=None, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment(
            [(0, 0), gap, (sw_dot_r*2, .1), (.8, .45), gap, (1, 0)]))
        self.segments.append(SegmentCircle((sw_dot_r, 0), sw_dot_r, fill='bg', zorder=3))
        self.segments.append(SegmentCircle((1-sw_dot_r, 0), sw_dot_r, fill='bg', zorder=3))
        if action == 'open':
            self.segments.append(SegmentArc((.4, .1), width=.5, height=.75,
                                            theta1=-10, theta2=70,
                                            arrow='ccw'))
        if action == 'close':
            self.segments.append(SegmentArc((.4, .25), width=.5, height=.75,
                                            theta1=-10, theta2=70,
                                            arrow='cw'))


class SwitchSpdt(Switch):
    ''' Single-pole double throw switch.

        Args:
            action: action arrow ('open' or 'close')

        Anchors:
            * a
            * b
            * c
    '''
    def __init__(self, *d, action: ActionType=None, **kwargs):
        super().__init__(*d, action=action, **kwargs)
        self.segments.append(SegmentCircle((1-sw_dot_r, .7), sw_dot_r, fill='bg', zorder=3))
        self.anchors['a'] = Point((sw_dot_r, 0))
        self.anchors['b'] = Point((1-sw_dot_r, 0))
        self.anchors['c'] = Point((1-sw_dot_r, .7))


class SwitchSpdt2(Element):
    ''' Single-pole double throw switch, throws above and below.

        Args:
            action: action arrow ('open' or 'close')

        Anchors:
            * a
            * b
            * c
    '''
    def __init__(self, *d, action: ActionType=None, **kwargs):
        super().__init__(*d, action=action, **kwargs)
        self.segments.append(Segment([(0, 0), gap, (sw_dot_r*2, .1),
                                      (.7, .25), gap, (1, .4)]))

        self.segments.append(SegmentCircle((sw_dot_r, 0), sw_dot_r, fill='bg', zorder=3))
        self.segments.append(SegmentCircle((1-sw_dot_r, -.4), sw_dot_r, fill='bg', zorder=3))
        self.segments.append(SegmentCircle((1-sw_dot_r, .4), sw_dot_r, fill='bg', zorder=3))
        self.anchors['a'] = Point((sw_dot_r, 0))
        self.anchors['b'] = Point((1-sw_dot_r, .4))
        self.anchors['c'] = Point((1-sw_dot_r, -.4))
        if action == 'open':
            self.segments.append(SegmentArc((.35, 0), width=.5, height=.75,
                                            theta1=-10, theta2=70,
                                            arrow='ccw'))
        elif action == 'close':
            self.segments.append(SegmentArc((.3, 0), width=.5, height=.75,
                                            theta1=-10, theta2=70,
                                            arrow='cw'))
        self.params['drop'] = (1, .4)


class Button(Element2Term):
    ''' Push button switch

        Args:
            nc: Normally closed
    '''
    def __init__(self, *d, nc: bool=False, **kwargs):
        super().__init__(*d, **kwargs)
        if nc:
            self.segments.append(Segment(
                [(0, 0), gap, (sw_dot_r, -sw_dot_r-.05),
                 (1-sw_dot_r, -sw_dot_r-.05), gap, (.5, -sw_dot_r-.05),
                 (.5, sw_dot_r+.15), gap, (1, 0)]))
        else:
            self.segments.append(Segment(
                [(0, 0), gap, (sw_dot_r, .3), (1-sw_dot_r, .3),
                 gap, (.5, .3), (.5, .5), gap, (1, 0)]))

        self.segments.append(SegmentCircle((sw_dot_r, 0), sw_dot_r))
        self.segments.append(SegmentCircle((1-sw_dot_r, 0), sw_dot_r))


class SwitchDpst(Element):
    ''' Double-pole single-throw switch

        Args:
            link: Show dotted line linking switch levers

        Anchors:
            * p1
            * p2
            * t1
            * t2
    '''
    def __init__(self, *d, link: bool=True, **kwargs):
        super().__init__(*d, **kwargs)
        yofst = -1
        self.segments.append(Segment([(0, 0), gap, (sw_dot_r*2, .1),
                                      (.8, .45), gap, (1, 0)]))
        self.segments.append(SegmentCircle((sw_dot_r, 0), sw_dot_r, fill='bg', zorder=3))
        self.segments.append(SegmentCircle((1-sw_dot_r, 0), sw_dot_r, fill='bg', zorder=3))
        self.segments.append(Segment([(0, yofst), gap, (sw_dot_r*2, yofst+.1),
                                      (.8, yofst+.45), gap, (1, yofst)]))
        self.segments.append(SegmentCircle((sw_dot_r, yofst), sw_dot_r, fill='bg', zorder=3))
        self.segments.append(SegmentCircle((1-sw_dot_r, yofst), sw_dot_r, fill='bg', zorder=3))
        if link:
            self.segments.append(Segment([(0.5, yofst+.25),
                                          (0.5, 0.2)], ls=':'))
        self.anchors['p1'] = (sw_dot_r, 0)
        self.anchors['t1'] = (1-sw_dot_r, 0)
        self.anchors['p2'] = (sw_dot_r, yofst)
        self.anchors['t2'] = (1-sw_dot_r, yofst)


class SwitchDpdt(Element):
    ''' Double-pole double-throw switch

        Args:
            link: Show dotted line linking switch levers

        Anchors:
            * p1
            * p2
            * t1
            * t2
            * t3
            * t4
    '''
    def __init__(self, *d, link: bool=True, **kwargs):
        super().__init__(*d, **kwargs)
        yofst = -1.4
        self.segments.append(Segment([(0, 0), gap, (sw_dot_r*2, .1),
                                      (.7, .25), gap, (1, .4)]))
        self.segments.append(SegmentCircle((sw_dot_r, 0), sw_dot_r, fill='bg', zorder=3))
        self.segments.append(SegmentCircle((1-sw_dot_r, -.4), sw_dot_r, fill='bg', zorder=3))
        self.segments.append(SegmentCircle((1-sw_dot_r, .4), sw_dot_r, fill='bg', zorder=3))
        self.segments.append(Segment([(0, yofst), gap, (sw_dot_r*2, yofst+.1),
                                      (.7, yofst+.25), gap, (1, yofst+.4)]))
        self.segments.append(SegmentCircle((sw_dot_r, yofst), sw_dot_r, fill='bg', zorder=3))
        self.segments.append(SegmentCircle((1-sw_dot_r, yofst-.4), sw_dot_r, fill='bg', zorder=3))
        self.segments.append(SegmentCircle((1-sw_dot_r, yofst+.4), sw_dot_r, fill='bg', zorder=3))
        if link:
            self.segments.append(Segment([(0.5, yofst+.25),
                                          (0.5, 0.2)], ls=':'))
        self.anchors['p1'] = (sw_dot_r, 0)
        self.anchors['t1'] = (1-sw_dot_r, .4)
        self.anchors['t2'] = (1-sw_dot_r, -.4)
        self.anchors['p2'] = (sw_dot_r, yofst)
        self.anchors['t3'] = (1-sw_dot_r, yofst+.4)
        self.anchors['t4'] = (1-sw_dot_r, yofst-.4)


class SwitchReed(Element2Term):
    ''' Reed Switch '''
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment(
            [(0, 0), (.85, .15), gap, (.8, 0)]))
        
        r = .3
        th = linspace(-math.pi/2, math.pi/2)
        x1 = [-r * math.cos(t) for t in th]  # Left semicircle
        y1 = [r * math.sin(t) for t in th]
        x2 = [1-k for k in x1]  # Right semicircle        

        x = x1 + x2 + [x1[0]]   # Combined
        y = y1 + y1[::-1] + [y1[0]]
        self.segments.append(Segment(list(zip(x, y))))

        
class SwitchRotary(Element):
    ''' Rotary Switch

        Args:
            n: number of contacts
            dtheta: angle in degrees between each contact
            theta0: angle in degrees of first contact
            radius: radius of switch
            arrowlen: length of switch arrow
            arrowcontact: index of contact to point to

        Values for dtheta and theta will be calculated based on `n`
        if not provided.
        
        Anchors:
            * P
            * T[x] for each contact (starting at 1)
    '''
    def __init__(self, *d, 
                 n: int=4, dtheta: float=None, theta0: float=None,
                 radius: float=1, arrowlen: float=0.75,
                 arrowcontact: int=0,
                 **kwargs):
        super().__init__(*d, **kwargs)
        self.params['fill'] = 'bg'
        self.params['zorder'] = 4
        self.segments.append(SegmentCircle((0, 0), sw_dot_r))
        self.anchors['P'] = (0, 0)

        if dtheta is None:
            dtheta = min(35, 360/(n+1))
        
        dtheta = math.radians(dtheta)
        if theta0 is None:
            theta0 = -dtheta * (n-1)/2

        for i in range(n):
            t = theta0 + dtheta * i
            x = radius * math.cos(t)
            y = radius * math.sin(t)
            self.segments.append(SegmentCircle((x, y), sw_dot_r))
            self.anchors[f'T{i+1}'] = (x, y)
        
            if i == arrowcontact:
                arrowx = arrowlen * math.cos(t)
                arrowy = arrowlen * math.sin(t)
                self.segments.append(Segment([(0, 0), (arrowx, arrowy)], arrow='->', zorder=2))


class SwitchDIP(Element):
    ''' DIP switch

        Args:
            n: Number of switches
            pattern: Boolean sequence indicating whether each switch is flipped up or down
            switchcolor: Fill color for flipped switches
            swidth: Width of one switch
            spacing: Spacing between switches
    '''
    def __init__(self, *d, n: int=3, pattern: Sequence[bool]=None, switchcolor:str ='#333333', 
                 swidth: float=0.4, spacing: float=0.2,
                 **kwargs):
        super().__init__(*d, **kwargs)
        
        width = swidth * n + spacing*(n+1)
        height = swidth*2 + spacing * 2
        
        self.segments.append(SegmentPoly(((0, 0), (width, 0), (width, height), (0, height))))
        for i in range(n):
            x = spacing * (i+1) + swidth * i
            up = pattern and pattern[i]
            down = pattern and not pattern[i]
            self.segments.append(SegmentPoly(((x, spacing+swidth), (x+swidth, spacing+swidth),
                                              (x+swidth, spacing+swidth*2), (x, spacing+swidth*2)),
                                              fill=switchcolor if up else None))  # Upper
            self.segments.append(SegmentPoly(((x, spacing), (x+swidth, spacing),
                                              (x+swidth, spacing+swidth), (x, spacing+swidth)),
                                              fill=switchcolor if down else None))  # Lower
            self.anchors[f'a{i+1}'] = (x + swidth/2, 0)
            self.anchors[f'b{i+1}'] = (x + swidth/2, height)


from .schemdraw import Drawing, use, config, theme
from .segments import Segment, SegmentCircle, SegmentArc, SegmentText, SegmentPoly, SegmentBezier, SegmentArrow
from .transform import Transform
from .types import ImageFormat
from .backends.svg import config as svgconfig
from .backends.svg import settextmode

__all__ = [
    "Drawing", "use", "config", "theme", "Segment", "SegmentCircle", "SegmentArc", "SegmentText",
    "SegmentPoly", "SegmentBezier", "Transform", "ImageFormat", "settextmode"
]

__version__ = '0.15'



''' Outlets/Plugs from various countries

    Reference: https://www.worldstandards.eu/electricity/plugs-and-sockets/
'''
import math

from schemdraw.segments import Segment, SegmentPoly, SegmentCircle
from schemdraw.elements import Element
from schemdraw.util import linspace, rotate

_outletrad = 1.1


class OutletA(Element):
    ''' Outlet Style A (Ungrounded, North America)

        Args:
            plug: Fill the prongs

        Anchors:
            * hot
            * neutral
    '''
    def __init__(self, *d, plug: bool=False, **kwargs):
        super().__init__(*d, **kwargs)
        self._outletrad = 1.5
        th = linspace(math.pi*3/4, math.pi*5/4)
        xleft = [.15 + self._outletrad * math.cos(t) for t in th]
        yleft = [self._outletrad * math.sin(t) for t in th]
        xrght = [-.15 + -self._outletrad * math.cos(t) for t in th]
        yrght = [-self._outletrad * math.sin(t) for t in th]
        x = xleft + xrght
        y = yleft + yrght
        self.segments.append(SegmentPoly(list(zip(x, y))))

        pinw = self._outletrad/8
        pinh = pinw*3
        pinhn = pinh*1.25
        pintop = pinh/2
        pintopn = pinhn/2
        pinleft = self._outletrad/2.2
        fill = 'black' if plug else 'bg'
        self.segments.append(SegmentPoly(((-pinleft, pintopn),
                                          (-pinleft, pintopn-pinhn),
                                          (-pinleft+pinw, pintopn-pinhn),
                                          (-pinleft+pinw, pintopn)),
                                         fill=fill, zorder=3))
        self.segments.append(SegmentPoly(((pinleft, pintop),
                                          (pinleft, pintop-pinh),
                                          (pinleft-pinw, pintop-pinh),
                                          (pinleft-pinw, pintop)),
                                         fill=fill, zorder=3))
        self.anchors['hot'] = (pinleft-pinw/2, 0)
        self.anchors['neutral'] = (-pinleft+pinw/2, 0)
        self.params['drop'] = (pinleft-pinw/2, 0)


class OutletB(OutletA):
    ''' Outlet Style B (Grounded, North America)

        Args:
            plug: Fill the prongs

        Anchors:
            * hot
            * neutral
            * ground
    '''
    def __init__(self, *d, plug: bool=False, **kwargs):
        super().__init__(*d, plug=plug, **kwargs)
        gndw = self._outletrad/4
        gndy = -self._outletrad/2.5
        x = [-gndw/2, -gndw/2, gndw/2, gndw/2]  # Flat part
        y = [gndy, gndy-gndw/2, gndy-gndw/2, gndy]
        theta = linspace(0, math.pi)
        xarc = [gndw/2 * math.cos(t) for t in theta]
        yarc = [gndy + gndw/2 * math.sin(t) for t in theta]
        x.extend(xarc)
        y.extend(yarc)
        fill = 'black' if plug else 'bg'
        self.segments.append(SegmentPoly((list(zip(x, y))), fill=fill, zorder=3))
        # Move prongs up a bit
        dy = gndw/2
        self.segments[1].verts = [(x, y+dy) for x, y in self.segments[1].verts]  # type: ignore
        self.segments[2].verts = [(x, y+dy) for x, y in self.segments[2].verts]  # type: ignore
        self.anchors['hot'] = self.anchors['hot'][0], self.anchors['hot'][1] + dy
        self.anchors['neutral'] = self.anchors['neutral'][0], self.anchors['neutral'][1] + dy
        self.anchors['ground'] = (0, gndy)


class OutletC(Element):
    ''' Outlet Style C (Ungrounded, Europe, South America, Asia)

        Args:
            plug: Fill the prongs

        Anchors:
            * hot
            * neutral
    '''
    def __init__(self, *d, plug: bool=False, **kwargs):
        super().__init__(*d, **kwargs)
        fill = 'black' if plug else 'bg'
        self.segments.append(SegmentCircle((0, 0), _outletrad))
        prad = _outletrad/8
        x = _outletrad/2
        self.segments.append(SegmentCircle((-x, 0), prad, fill=fill, zorder=3))
        self.segments.append(SegmentCircle((x, 0), prad, fill=fill, zorder=3))
        self.anchors['hot'] = (x, 0)
        self.anchors['neutral'] = (-x, 0)
        self.params['drop'] = self.anchors['hot']

class OutletD(Element):
    ''' Outlet Style D (India)

        Args:
            plug: Fill the prongs

        Anchors:
            * hot
            * neutral
            * ground
    '''
    def __init__(self, *d, plug: bool=False, **kwargs):
        super().__init__(*d, **kwargs)
        fill = 'black' if plug else 'bg'
        self.segments.append(SegmentCircle((0, 0), _outletrad))
        prad = _outletrad/8
        x = _outletrad/2
        y = -_outletrad/4
        y2 = _outletrad/2.5
        self.segments.append(SegmentCircle((-x, y), prad, fill=fill, zorder=3))
        self.segments.append(SegmentCircle((x, y), prad, fill=fill, zorder=3))
        self.segments.append(SegmentCircle((0, y2), prad*1.2, fill=fill, zorder=3))
        self.anchors['hot'] = (x, y)
        self.anchors['neutral'] = (-x, y)
        self.anchors['ground'] = (0, y2)
        self.params['drop'] = self.anchors['ground']


class OutletE(Element):
    ''' Outlet Style D (France, Belgium, Poland, ...)

        Args:
            plug: Fill the prongs

        Anchors:
            * hot
            * neutral
            * ground
    '''
    def __init__(self, *d, plug: bool=False, **kwargs):
        super().__init__(*d, **kwargs)
        fill = 'black' if plug else 'bg'
        self.segments.append(SegmentCircle((0, 0), _outletrad))
        prad = _outletrad/8
        x = _outletrad/2
        y = 0
        y2 = _outletrad/2
        self.segments.append(SegmentCircle((-x, y), prad, fill=fill, zorder=3))
        self.segments.append(SegmentCircle((x, y), prad, fill=fill, zorder=3))
        self.segments.append(SegmentCircle((0, y2), prad, fill=fill, zorder=3))
        self.anchors['hot'] = (x, y)
        self.anchors['neutral'] = (-x, y)
        self.anchors['ground'] = (0, y2)
        self.params['drop'] = self.anchors['neutral']


class OutletF(OutletC):
    ''' Outlet Style F (Europe, Russia)

        Args:
            plug: Fill the prongs

        Anchors:
            * hot
            * neutral
    '''
    def __init__(self, *d, plug: bool=False, **kwargs):
        super().__init__(*d, plug=plug, **kwargs)
        fill = 'black' if plug else 'bg'
        notchw = _outletrad/8
        notchH = _outletrad/6
        y = _outletrad - .05
        y2 = -_outletrad + .05
        self.segments.append(Segment(((-notchw, y), (-notchw, y-notchH),
                                      (notchw, y-notchH), (notchw, y)),
                                     fill=fill, zorder=3))
        self.segments.append(Segment(((-notchw, y2), (-notchw, y2+notchH),
                                      (notchw, y2+notchH), (notchw, y2)),
                                     fill=fill, zorder=3))


class OutletG(Element):
    ''' Outlet Style G (U.K.)

        Args:
            plug: Fill the prongs

        Anchors:
            * hot
            * neutral
            * ground
    '''
    def __init__(self, *d, plug: bool=False, **kwargs):
        super().__init__(*d, **kwargs)
        fill = 'black' if plug else 'bg'
        self.segments.append(SegmentCircle((0, 0), _outletrad))
        pinw = _outletrad/6
        pinh = pinw*3
        gndw = _outletrad/5
        gndtop = _outletrad*3/4
        y = -_outletrad/2.5
        pinleft = pinw
        fill = 'black' if plug else 'bg'
        self.segments.append(SegmentPoly(((-gndw/2, gndtop), (-gndw/2, gndtop-pinh),
                                          (gndw/2, gndtop-pinh), (gndw/2, gndtop)),
                                        fill=fill, zorder=3))
        self.segments.append(SegmentPoly(((-pinleft, y), (-pinleft, y-pinw),
                                          (-pinleft-pinh, y-pinw), (-pinleft-pinh, y)),
                                         fill=fill, zorder=3))
        self.segments.append(SegmentPoly(((pinleft, y), (pinleft, y-pinw),
                                          (pinleft+pinh, y-pinw), (pinleft+pinh, y)),
                                         fill=fill, zorder=3))
        self.anchors['hot'] = (pinleft+pinh/2, y-pinw/2)
        self.anchors['neutral'] = (-pinleft-pinh/2, y-pinw/2)
        self.anchors['ground'] = (0, gndtop-pinh/2)
        self.params['drop'] = self.anchors['hot']


class OutletH(Element):
    ''' Outlet Style H (Israel)

        Args:
            plug: Fill the prongs

        Anchors:
            * hot
            * neutral
            * ground
    '''
    def __init__(self, *d, plug: bool=False, **kwargs):
        super().__init__(*d, **kwargs)
        fill = 'black' if plug else 'bg'
        self.segments.append(SegmentCircle((0, 0), _outletrad))
        pinw = _outletrad/8
        pinh = pinw*3
        gndtop = -pinw*2
        verts = ((-pinw/2, gndtop), (-pinw/2, gndtop-pinh), (pinw/2, gndtop-pinh), (pinw/2, gndtop))
        hotverts = [rotate(v, 120) for v in verts]
        neutverts = [rotate(v, -120) for v in verts]
        self.segments.append(SegmentPoly(verts, fill=fill, zorder=3))
        self.segments.append(SegmentPoly(hotverts, fill=fill, zorder=3))
        self.segments.append(SegmentPoly(neutverts, fill=fill, zorder=3))
        self.anchors['hot'] = (hotverts[1])
        self.anchors['neutral'] = (neutverts[2])
        self.anchors['ground'] = (0, gndtop-pinh)
        self.params['drop'] = self.anchors['hot']


class OutletI(OutletH):
    ''' Outlet Style I (Australia, New Zealand, China)

        Args:
            plug: Fill the prongs

        Anchors:
            * hot
            * neutral
            * ground
    '''
    def __init__(self, *d, plug: bool=False, **kwargs):
        super().__init__(*d, plug=plug, **kwargs)
        hotcenter = (.4, .3)
        neutcenter = (-.4, .3)
        self.segments[2].verts = [rotate(v, 90, center=hotcenter) for v in self.segments[2].verts]  # type: ignore
        self.segments[3].verts = [rotate(v, -90, center=neutcenter) for v in self.segments[3].verts]  # type: ignore
        self.anchors['hot'] = rotate(self.anchors['hot'], 90, center=hotcenter)
        self.anchors['neutral'] = rotate(self.anchors['neutral'], -90, center=neutcenter)


class OutletJ(Element):
    ''' Outlet Style J (Switzerland)

        Args:
            plug: Fill the prongs

        Anchors:
            * hot
            * neutral
            * ground
    '''
    def __init__(self, *d, plug: bool=False, **kwargs):
        super().__init__(*d, **kwargs)
        fill = 'black' if plug else 'bg'
        fullh = _outletrad*1.5
        fullw = fullh*2
        dw = fullw/5
        self.segments.append(SegmentPoly(((-fullw/2, 0), (-fullw/2+dw, fullh/2),
                                          (fullw/2-dw, fullh/2), (fullw/2, 0),
                                          (fullw/2-dw, -fullh/2),
                                          (-fullw/2+dw, -fullh/2))))
        prad = _outletrad/8
        self.segments.append(SegmentCircle((-fullw/4, 0), prad, fill=fill, zorder=3))
        self.segments.append(SegmentCircle((0, -prad*2), prad, fill=fill, zorder=3))
        self.segments.append(SegmentCircle((fullw/4, 0), prad, fill=fill, zorder=3))
        self.anchors['hot'] = (fullw/4, 0)
        self.anchors['neutral'] = (-fullw/4, 0)
        self.anchors['ground'] = (0, -prad*2)
        self.params['drop'] = self.anchors['hot']


class OutletK(Element):
    ''' Outlet Style K (Denmark, Greenland)

        Args:
            plug: Fill the prongs

        Anchors:
            * hot
            * neutral
            * ground
    '''
    def __init__(self, *d, plug: bool=False, **kwargs):
        super().__init__(*d, **kwargs)
        fill = 'black' if plug else 'bg'
        self.segments.append(SegmentCircle((0, 0), _outletrad))
        prad = _outletrad/8
        prongx = _outletrad/2
        self.segments.append(SegmentCircle((-prongx, 0), prad, fill=fill, zorder=3))
        self.segments.append(SegmentCircle((prongx, 0), prad, fill=fill, zorder=3))
        self.anchors['hot'] = (prongx, 0)
        self.anchors['neutral'] = (-prongx, 0)
        self.params['drop'] = (prongx, 0)
        gndw = _outletrad/4
        gndy = -_outletrad/2
        x = [-gndw/2, -gndw/2, gndw/2, gndw/2]  # Flat part
        y = [gndy, gndy+gndw/2, gndy+gndw/2, gndy]
        theta = linspace(0, math.pi)
        xarc = [gndw/2 * math.cos(t) for t in theta]
        yarc = [gndy - gndw/2 * math.sin(t) for t in theta]
        x.extend(xarc)
        y.extend(yarc)
        self.segments.append(SegmentPoly(list(zip(x, y)), fill=fill, zorder=3))
        self.anchors['ground'] = (0, gndy)


class OutletL(Element):
    ''' Outlet Style L (Italy, Chile)

        Args:
            plug: Fill the prongs

        Anchors:
            * hot
            * neutral
            * ground
    '''
    def __init__(self, *d, plug: bool=False, **kwargs):
        super().__init__(*d, **kwargs)
        fill = 'black' if plug else 'bg'
        fullh = _outletrad*1.2
        fullw = fullh*2.25
        self.segments.append(SegmentPoly(((-fullw/2, fullh/2), (fullw/2, fullh/2),
                                          (fullw/2, -fullh/2), (-fullw/2, -fullh/2))))
        prad = _outletrad/8
        self.segments.append(SegmentCircle((-fullw/4, 0), prad, fill=fill, zorder=3))
        self.segments.append(SegmentCircle((0, 0), prad, fill=fill, zorder=3))
        self.segments.append(SegmentCircle((fullw/4, 0), prad, fill=fill, zorder=3))
        self.anchors['hot'] = (fullw/4, 0)
        self.anchors['neutral'] = (-fullw/4, 0)
        self.anchors['ground'] = (0, 0)
        self.params['drop'] = self.anchors['hot']
